! ---------------------------------------------------------------------------
! L_tL_parallel_test6.f90  - by Alejandro Valenciano 
!                        
!           March  2005.
! ---------------------------------------------------------------------------
!       Usage: 
!              L_tL_parallel_test6.x  
!  Parameters: 
!              nxh     x offset to calculate the Hessian (odd)
!              nzh     z offset to calculate the Hessian (odd)
!              tsid    true if receiver at each side of the shot
!         I/O:
!              in      green_in function (nsx,nz,nx) 
!              out     Hessian           (nz,nx_out,nhz,nhx,w) 
!                          
! Description: 
!              Computes the Hessian 
!              Parallelizable in the omega dimension
! ---------------------------------------------------------------------------

program L_tL_para 
use sep

implicit none

integer                                  :: n1,n2,n3,n4
integer                                  :: nsx,nz,nx,nw,nxh,nzh
real                                     :: o1,o2,o3,o4
real                                     :: d1,d2,d3,d4 
integer                                  :: esize,iz,ix,ihz,ihx,isr,iw,istart,i1
real                                     :: mini,maxi,start
integer                                  :: tsid
complex, dimension(:,:,:)  , allocatable :: green,green_in
complex, dimension(:)      , allocatable :: aaaixz,aaajxz 
complex, dimension(:,:)    , allocatable :: tmp 
real,    dimension(:,:,:,:), allocatable :: LtL 
complex, dimension(:,:,:,:), allocatable :: MulGreen 

call sep_init()
call from_history('n1',n1); call from_history('d1',d1); call from_history('o1',o1)
call from_history('n2',n2); call from_history('d2',d2); call from_history('o2',o2)
call from_history('n3',n3); call from_history('d3',d3); call from_history('o3',o3)
call from_history('n4',n4); call from_history('d4',d4); call from_history('o4',o4)
call from_history("esize",esize)

call from_param('nxh',nxh,11)        ! allways odd 
call from_param('nzh',nzh,11)        ! allways odd
call from_param('tsid',tsid,1)       ! 1 if receiver at each side of the shot
if (tsid==1) then                    ! same # receiver at each side of the shot
  mini=o3+o1
else                                 ! receiver at the rigth of the shot 
  mini=o3                             ! assumes o3=0 first offset=0 
endif
maxi=mini+(n3-1)*d3+(n1-1)*d1
nx=1.5+(maxi-mini)/d1                  ! new nx
 
nsx=n3
nz=n2
nw=n4
call to_history('n1',nz ); call to_history('d1',d2); call to_history('o1',o2) ; call to_history('label1',"z")
call to_history('n2',nx ); call to_history('d2',d1); call to_history('o2',mini) ; call to_history('label2',"mx")
call to_history('n3',nzh); call to_history('d3',d2); call to_history('o3',0.) ; call to_history('label3',"hz")
call to_history('n4',nxh); call to_history('d4',d1); call to_history('o4',0.) ; call to_history('label4',"hx")
call to_history('n5',nw ); call to_history('d5',d4); call to_history('o5',o4) ; call to_history('label5',"w")
call to_history('esize',4)
call sep_close()

call sep_begin_prog()

write(0,*)"nsx,nz,nx,nw",nsx,nz,nx,nw
write(0,*)"d1,d2,d3,d4",d1,d2,d3,n4
write(0,*)"o1,o2,o3,o4",o1,o2,o3,n4

allocate(green_in(n1,nz,nsx),tmp(nz,n1),green(nsx,nz,nx),LtL(nz,nx,nzh,nxh), &
         MulGreen(nz,nx,nzh,nxh),aaaixz(nsx),aaajxz(nsx))

do iw=1,nw
   LtL=0.;
   MulGreen=(0.,0.)
   green=(0.,0.)
   write(0,*)"frequency",iw
   call sep_read(green_in)

   do isr=1,nsx
     start=mini+(isr-1)*d3
     istart=1.5+(start-mini)/d1
     do iz=1,nz
     do i1=1,n1
       green(isr,iz,istart+i1-1)=green_in(i1,iz,isr)
     end do
     end do
   end do

   ! Computation of the second Sum
   do ix=1,nx
   do iz=1,nz
     aaaixz(1:nsx)=conjg(green(1:nsx,iz,ix))
     do ihz=1,nzh
     do ihx=1,nxh
       if (  (iz-(nzh-1)/2+ihz-1 >= 1  ) .and. (ix-(nxh-1)/2+ihx-1 >=1  ) .and. &
             (iz-(nzh-1)/2+ihz-1 <= nz ) .and. (ix-(nxh-1)/2+ihx-1 <= nx)) then
         aaajxz(1:nsx)=green(1:nsx,iz-(nzh-1)/2+ihz-1,ix-(nxh-1)/2+ihx-1)          !  G(h_z,h_x)
       else
         aaajxz(1:nsx)=(0.,0.)
       endif
       do isr=1,nsx
         MulGreen(iz,ix,ihz,ihx)=MulGreen(iz,ix,ihz,ihx)+aaaixz(isr)*aaajxz(isr)
       enddo
     enddo
     enddo
   enddo
   enddo
   
   ! Computation of the Hessian
   do ix=1,nx
   do iz=1,nz
     aaaixz(1:nsx)=conjg(green(1:nsx,iz,ix)) 
     do ihz=1,nzh
     do ihx=1,nxh
       if (  (iz-(nzh-1)/2+ihz-1 >= 1  ) .and. (ix-(nxh-1)/2+ihx-1 >=1  ) .and. &
             (iz-(nzh-1)/2+ihz-1 <= nz ) .and. (ix-(nxh-1)/2+ihx-1 <= nx)) then
         aaajxz(1:nsx)=green(1:nsx,iz-(nzh-1)/2+ihz-1,ix-(nxh-1)/2+ihx-1)
       else
         aaajxz(1:nsx)=(0.,0.)
       endif
       do isr=1,nsx
         LtL(iz,ix,ihz,ihx)=LtL(iz,ix,ihz,ihx)+real( aaaixz(isr)*aaajxz(isr)*MulGreen(iz,ix,ihz,ihx) )
       enddo
     enddo
     enddo
   enddo
   enddo
   call sep_write(LtL)
enddo

deallocate(green_in,tmp,green,LtL,MulGreen,aaaixz,aaajxz)

call sep_end_prog()

end program L_tL_para
